package org.chulk.codegen;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.charset.Charset;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.chulk.codegen.command.construct.ModelValue;
import org.chulk.codegen.command.construct.PayloadElement;
import org.chulk.codegen.command.construct.StringLiteral;
import org.joor.Reflect;

import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.base.Splitter;
import com.google.common.collect.Lists;

import lombok.extern.java.Log;

@Log
public class CodeGenerator {

	private final Charset charset;

	private CodeGenerator(final Charset charset) {
		this.charset = charset;
	}

	public static CodeGenerator create(final Charset charset) {
		return new CodeGenerator(charset);
	}

	/**
	 * Write generator code.
	 * 
	 * <p>
	 * Takes the template file at the templatePath and outputs Java generator code
	 *
	 * @param templatePath the template path
	 * @return the string
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	public String writeGeneratorCode(final String templatePath) throws IOException {
		try (final BufferedReader reader = new BufferedReader(
				new InputStreamReader(new FileInputStream(templatePath), this.charset))) {

			String lineToProcess = "";
			StringBuilder strBuilder = new StringBuilder();

			while (reader.ready()) {
				lineToProcess += reader.readLine();

				if (lineToProcess.trim().matches("<<:(out|method)\\s.*>>")) { //unary function
					buildJavaCode(lineToProcess, strBuilder, 2);
					lineToProcess = "";
				} else if (lineToProcess.trim().matches("<<:(file)\\s.*>>")) { //binary function
					buildJavaCode(lineToProcess, strBuilder, 3);
					lineToProcess = "";
				} else if (lineToProcess.trim().matches("<<:(begin|end).*>>")) {
					String lineContent = lineToProcess.trim().replaceFirst("^<<:", "").replaceFirst(">>$", "");
					strBuilder.append(templateCommandToJava(lineContent, null, null));

					lineToProcess = "";
				} else if (lineToProcess.trim().startsWith("<<:out")) { // non-terminal out function
					log.info("Non-terminal out command");
					lineToProcess += " \\n";
				} else {
					strBuilder.append(lineToProcess + "\n");
					lineToProcess = "";
				}
			}

			return strBuilder.toString();
		}
	}

	private void buildJavaCode(String lineToProcess, StringBuilder strBuilder, int payloadIndex) {
		String lineContent = lineToProcess.trim().replaceFirst("^<<:", "").replaceFirst(">>$", "");
		List<String> contentTokens = Splitter.on(" ").trimResults().splitToList(lineContent);
		String payload = Joiner.on(" ").join(contentTokens.subList(payloadIndex, contentTokens.size()));

		strBuilder.append(templateCommandToJava(contentTokens.get(0), collectArgs(contentTokens, payloadIndex), evalPayload(payload))
				+ "\n");
	}
	
	private List<String> collectArgs(final List<String> contentTokens, final int argEndIndex) {
		List<String> args = Lists.newArrayList();
		
		for (int n = 1; n < argEndIndex; n++) {
			args.add(contentTokens.get(n));
		}
		
		return args;
	}

	private String templateCommandToJava(String fnName, List<String> args, final Queue<PayloadElement> queue) {
		Preconditions.checkArgument(Command.valueOf(fnName) != null, "Non-supported command");

		return Command.valueOf(fnName).getMapper().map(args, queue);
	}

	private Queue<PayloadElement> evalPayload(final String payload) {
		final char[] c = payload.toCharArray();
		final Queue<PayloadElement> payloadQueue = new LinkedList<>();
		boolean isModelValue = false;
		String tmpToken = "";

		for (int n = 0; n < c.length; n++) {
			if (c[n] == '{') {
				addPayloadElement(payloadQueue, isModelValue, tmpToken);
				isModelValue = true;
				tmpToken = "";
			} else if (c[n] == '}') {
				addPayloadElement(payloadQueue, isModelValue, tmpToken);
				isModelValue = false;
				tmpToken = "";
			} else {
				tmpToken += c[n];
			}
		}

		if (!tmpToken.isEmpty()) {
			addPayloadElement(payloadQueue, isModelValue, tmpToken);
		}

		return payloadQueue;
	}

	private void addPayloadElement(final Queue<PayloadElement> payloadQueue, final boolean isModelValue,
			final String tmpToken) {
		PayloadElement payloadElement = isModelValue ? new ModelValue(tmpToken) : new StringLiteral(tmpToken);
		payloadQueue.add(payloadElement);
	}

	/**
	 * Generate code.
	 *
	 * @param <T>       the generic type
	 * @param packagePath the package path
	 * @param content   the content
	 * @param model     the model
	 * @throws NoSuchMethodException     the no such method exception
	 * @throws IllegalAccessException    the illegal access exception
	 * @throws InvocationTargetException the invocation target exception
	 */
	public <T> void generateCode(final String packagePath, final String content, final T model) //Add type.. as Model class can be of a super class
			throws NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		final Object genObj = Reflect.compile(packagePath, content).create().get();
		final Method genMethod = genObj.getClass().getDeclaredMethods()[0];
		genMethod.invoke(genObj, model);
		
	}

}
